@@ -1,6 +1,6 @@
 /* 
    HTTP Authentication routines
-   Copyright (C) 1999-2001, Joe Orton <joe@light.plus.com>
+   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -23,27 +23,34 @@
 /* HTTP Authentication, as per RFC2617.
  * 
  * TODO:
- *  - Improve cnonce? Make it really random using /dev/random or whatever?
  *  - Test auth-int support
  */
 
 #include "config.h"
 
+#include <sys/types.h>
+
+#ifdef HAVE_SYS_TIME_H
+#include <sys/time.h>
+#endif
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-
-#include <time.h>
-
-#ifdef HAVE_SNPRINTF_H
-#include "snprintf.h"
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
+#endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h> /* for getpid() */
 #endif
 
-#include "base64.h"
+#ifdef NEON_SSL
+#include <openssl/rand.h>
+#endif
+
+#include <time.h>
 
 #include "ne_md5.h"
 #include "ne_dates.h"
@@ -55,6 +62,15 @@
 #include "ne_uri.h"
 #include "ne_i18n.h"
 
+#ifdef HAVE_GSSAPI
+#ifdef HAVE_GSSAPI_GSSAPI_H /* MIT */
+#include <gssapi/gssapi.h>
+#include <gssapi/gssapi_generic.h>
+#else /* Heimdal. */
+#include <gssapi.h>
+#endif
+#endif
+
 /* TODO: should remove this eventually. Need it for
  * ne_pull_request_body. */
 #include "ne_private.h"
@@ -65,7 +81,8 @@
 /* The authentication scheme we are using */
 typedef enum {
     auth_scheme_basic,
-    auth_scheme_digest
+    auth_scheme_digest,
+    auth_scheme_gssapi
 } auth_scheme;
 
 typedef enum { 
@@ -85,7 +102,6 @@ typedef enum {
 struct auth_challenge {
     auth_scheme scheme;
     const char *realm;
-    const char *domain;
     const char *nonce;
     const char *opaque;
     unsigned int stale:1; /* if stale=true */
@@ -96,56 +112,50 @@ struct auth_challenge {
     struct auth_challenge *next;
 };
 
-static const char *qop_values[] = {
-    NULL,
-    "auth",
-    "auth-int"
+static const struct auth_class {
+    const char *id, *req_hdr, *resp_hdr, *resp_info_hdr, *fail_msg;
+    int status_code, fail_code;
+} ah_server_class = {
+    HOOK_SERVER_ID,
+    "Authorization", "WWW-Authenticate", "Authentication-Info",
+    N_("Server was not authenticated correctly."), 401, NE_AUTH
+}, ah_proxy_class = {
+    HOOK_PROXY_ID,
+    "Proxy-Authorization", "Proxy-Authenticate", "Proxy-Authentication-Info",
+    N_("Proxy server was not authenticated correctly."), 407, NE_PROXYAUTH 
 };
-static const char *algorithm_names[] = {
-    "MD5",
-    "MD5-sess",
-    NULL
-};
-
-/* The callback used to request the username and password in the given
- * realm. The username and password must be placed in malloc()-allocate
- * memory.
- * Must return:
- *   0 on success, 
- *  -1 to cancel.
- */
 
 /* Authentication session state. */
 typedef struct {
     ne_session *sess;
-
-    /* for making this proxy/server generic. */
-    const char *req_hdr, *resp_hdr, *resp_info_hdr, *fail_msg;
-    int status_code;
-    int fail_code;
-
+    
+    /* Specifics for server/proxy auth. FIXME: need a better field
+     * name! */
+    const struct auth_class *spec;
+    
     /* The scheme used for this authentication session */
     auth_scheme scheme;
     /* The callback used to request new username+password */
-    ne_request_auth reqcreds;
-    void *reqcreds_udata;
+    ne_auth_creds creds;
+    void *userdata;
 
     /*** Session details ***/
 
     /* The username and password we are using to authenticate with */
-    char *username;
+    char username[NE_ABUFSIZ];
     /* Whether we CAN supply authentication at the moment */
     unsigned int can_handle:1;
     /* This used for Basic auth */
     char *basic; 
+#ifdef HAVE_GSSAPI
+    /* This used for GSSAPI auth */
+    char *gssapi_token;
+#endif
     /* These all used for Digest auth */
     char *realm;
     char *nonce;
     char *cnonce;
     char *opaque;
-    /* A list of domain strings */
-    unsigned int domain_count;
-    char **domain;
     auth_qop qop;
     auth_algorithm alg;
     int nonce_count;
@@ -162,11 +172,10 @@ typedef struct {
     const char *uri_scheme;
     unsigned int port;
 
+    int attempt;
 } auth_session;
 
 struct auth_request {
-    auth_session *session;
-
     /*** Per-request details. ***/
     ne_request *request; /* the request object. */
 
@@ -179,68 +188,10 @@ struct auth_request {
     /* Used for calculation of H(entity-body) of the response */
     struct ne_md5_ctx response_body;
 
-    int tries;
     /* Results of response-header callbacks */
     char *auth_hdr, *auth_info_hdr;
 };
 
-/* Private prototypes which used to be public. */
-
-static auth_session *auth_create(ne_session *sess,
-				 ne_request_auth callback, 
-				 void *userdata);
-
-/* Pass this the value of the "(Proxy,WWW)-Authenticate: " header field.
- * Returns:
- *   0 if we can now authenticate ourselves with the server.
- *   non-zero if we can't
- */
-static int auth_challenge(auth_session *sess, const char *value);
-
-/* If you receive a "(Proxy-)Authentication-Info:" header, pass its value to
- * this function. Returns zero if this successfully authenticates
- * the response as coming from the server, and false if it hasn't. */
-static int verify_response(struct auth_request *req, 
-			   auth_session *sess, const char *value);
-
-/* Private prototypes */
-static char *get_cnonce(void);
-static void clean_session(auth_session *sess);
-static int digest_challenge(auth_session *, struct auth_challenge *);
-static int basic_challenge(auth_session *, struct auth_challenge *);
-static char *request_digest(auth_session *sess, struct auth_request *req);
-static char *request_basic(auth_session *);
-
-/* Domain handling */
-static int is_in_domain(auth_session *sess, const char *uri);
-static int parse_domain(auth_session *sess, const char *domain);
-
-/* Get the credentials, passing a temporary store for the password value */
-static int get_credentials(auth_session *sess, char **password);
-
-auth_session *auth_create(ne_session *sess,
-			  ne_request_auth callback, 
-			  void *userdata) 
-{
-    auth_session *ret = ne_calloc(sizeof(auth_session));
-
-    ret->reqcreds = callback;
-    ret->reqcreds_udata = userdata;
-    ret->sess = sess;
-   
-    return ret;
-}
-
-#if 0
-void ne_auth_set_server(auth_session *sess, 
-			   const char *host, unsigned int port, const char *scheme)
-{
-    sess->host = host;
-    sess->port = port;
-    sess->req_scheme = scheme;
-}
-#endif
-
 static void clean_session(auth_session *sess) 
 {
     sess->can_handle = 0;
@@ -248,88 +199,75 @@ static void clean_session(auth_session *sess)
     NE_FREE(sess->nonce);
     NE_FREE(sess->cnonce);
     NE_FREE(sess->opaque);
-    NE_FREE(sess->username);
     NE_FREE(sess->realm);
-    if (sess->domain_count > 0) {
-	split_string_free(sess->domain);
-	sess->domain_count = 0;
-    }
+#ifdef HAVE_GSSAPI
+    NE_FREE(sess->gssapi_token);
+#endif
 }
 
-/* Returns cnonce-value. We just use base64(time).
- * TODO: Could improve this? */
+/* Returns client nonce string. */
 static char *get_cnonce(void) 
 {
-    char *ret, *tmp;
-    tmp = ne_rfc1123_date(time(NULL));
-    ret = ne_base64(tmp, strlen(tmp));
-    free(tmp);
-    return ret;
-}
+    char data[256], ret[33];
+    unsigned char tmp[16];
+    struct ne_md5_ctx hash;
 
-static int 
-get_credentials(auth_session *sess, char **password) 
-{
-    return (*sess->reqcreds)(sess->reqcreds_udata, sess->realm,
-			     &sess->username, password);
-}
+    ne_md5_init_ctx(&hash);
 
-static int parse_domain(auth_session *sess, const char *domain) {
-    char **doms;
-    int count, ret;
-
-    doms = split_string_c(domain, ' ', NULL, " \r\n\t", &count);
-    if (doms != NULL) {
-	if (count > 0) {
-	    sess->domain = doms;
-	    sess->domain_count = count;
-	    ret = 0;
-	} else {
-	    free(doms);
-	    ret = -1;
-	}
-    } else {
-	ret = -1;
-    }
-
-    return ret;
-}
-
-/* Returns:
- *    0: if uri is in NOT in domain of session
- * else: uri IS in domain of session (or no domain known)
- */
-static int is_in_domain(auth_session *sess, const char *uri)
-{
-#if 1
-    return 1;
-#else
-    /* TODO: Need proper URI comparison for this to work. */
-    int ret, dom;
-    const char *abs_path;
-    if (sess->domain_count == 0) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "No domain, presuming okay.\n");
-	return 1;
-    }
-    ret = 0;
-    abs_path = uri_abspath(uri);
-    for (dom = 0; dom < sess->domain_count; dom++) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Checking domain: %s\n", sess->domain[dom]);
-	if (uri_childof(sess->domain[dom], abs_path)) {
-	    ret = 1;
-	    break;
-	}
-    }
-    return ret;
+#ifdef NEON_SSL
+    if (RAND_pseudo_bytes(data, sizeof data) >= 0)
+	ne_md5_process_bytes(data, sizeof data, &hash);
+    else {
 #endif
+    /* Fallback sources of random data: all bad, but no good sources
+     * are available. */
+
+    /* Uninitialized stack data; yes, happy valgrinders, this is
+     * supposed to be here. */
+    ne_md5_process_bytes(data, sizeof data, &hash);
+    
+#ifdef HAVE_GETTIMEOFDAY
+    {
+	struct timeval tv;
+	if (gettimeofday(&tv, NULL) == 0)
+	    ne_md5_process_bytes(&tv, sizeof tv, &hash);
+    }
+#else /* HAVE_GETTIMEOFDAY */
+    {
+	time_t t = time(NULL);
+	ne_md5_process_bytes(&t, sizeof t, &hash);
+    }
+#endif
+    {
+#ifdef WIN32
+	DWORD pid = GetCurrentThreadId();
+#else
+	pid_t pid = getpid();
+#endif
+	ne_md5_process_bytes(&pid, sizeof pid, &hash);
+    }
+
+#ifdef NEON_SSL
+    }
+#endif
+    
+    ne_md5_finish_ctx(&hash, tmp);
+    ne_md5_to_ascii(tmp, ret);
+
+    return ne_strdup(ret);
+}
+
+static int get_credentials(auth_session *sess, char *pwbuf) 
+{
+    return sess->creds(sess->userdata, sess->realm, sess->attempt++,
+		       sess->username, pwbuf);
 }
 
 /* Examine a Basic auth challenge.
  * Returns 0 if an valid challenge, else non-zero. */
-static int 
-basic_challenge(auth_session *sess, struct auth_challenge *parms) 
+static int basic_challenge(auth_session *sess, struct auth_challenge *parms) 
 {
-    char *tmp, *password;
+    char *tmp, password[NE_ABUFSIZ];
 
     /* Verify challenge... must have a realm */
     if (parms->realm == NULL) {
@@ -337,24 +275,25 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
     }
 
     NE_DEBUG(NE_DBG_HTTPAUTH, "Got Basic challenge with realm [%s]\n", 
-	   parms->realm);
+	     parms->realm);
 
     clean_session(sess);
     
     sess->realm = ne_strdup(parms->realm);
 
-    if (get_credentials(sess, &password)) {
+    if (get_credentials(sess, password)) {
 	/* Failed to get credentials */
 	return -1;
     }
 
     sess->scheme = auth_scheme_basic;
 
-    CONCAT3(tmp, sess->username, ":", password?password:"");
+    tmp = ne_concat(sess->username, ":", password, NULL);
     sess->basic = ne_base64(tmp, strlen(tmp));
-    free(tmp);
+    ne_free(tmp);
 
-    NE_FREE(password);
+    /* Paranoia. */
+    memset(password, 0, sizeof password);
 
     return 0;
 }
@@ -362,19 +301,83 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
 /* Add Basic authentication credentials to a request */
 static char *request_basic(auth_session *sess) 
 {
-    char *buf;
-    CONCAT3(buf, "Basic ", sess->basic, "\r\n");
-    return buf;
+    return ne_concat("Basic ", sess->basic, "\r\n", NULL);
 }
 
+#ifdef HAVE_GSSAPI
+/* Add GSSAPI authentication credentials to a request */
+static char *request_gssapi(auth_session *sess) 
+{
+    return ne_concat("GSS-Negotiate ", sess->gssapi_token, "\r\n", NULL);
+}
+
+static int get_gss_name(gss_name_t *server, auth_session *sess)
+{
+    int major_status, minor_status;
+    gss_buffer_desc token = GSS_C_EMPTY_BUFFER;
+
+    token.value = ne_concat("khttp@", sess->sess->server.hostname, NULL);
+    token.length = strlen(token.value);
+
+    major_status = gss_import_name(&minor_status, &token,
+                                   GSS_C_NT_HOSTBASED_SERVICE,
+                                   server);
+    return GSS_ERROR(major_status) ? -1 : 0;
+}
+
+/* Examine a GSSAPI auth challenge; returns 0 if a valid challenge,
+ * else non-zero. */
+static int 
+gssapi_challenge(auth_session *sess, struct auth_challenge *parms) 
+{
+    gss_ctx_id_t context;
+    gss_name_t server_name;
+    int major_status, minor_status;
+    gss_buffer_desc input_token = GSS_C_EMPTY_BUFFER;
+    gss_buffer_desc output_token = GSS_C_EMPTY_BUFFER;
+
+    clean_session(sess);
+
+    if (get_gss_name(&server_name, sess))
+        return -1;
+
+    major_status = gss_init_sec_context(&minor_status,
+                                        GSS_C_NO_CREDENTIAL,
+                                        &context,
+                                        server_name,
+                                        GSS_C_NO_OID,
+                                        GSS_C_DELEG_FLAG,
+                                        0,
+                                        GSS_C_NO_CHANNEL_BINDINGS,
+                                        &input_token,
+                                        NULL,
+                                        &output_token,
+                                        NULL,
+                                        NULL);
+    if (GSS_ERROR(major_status)) {
+        NE_DEBUG(NE_DBG_HTTPAUTH, "gss_init_sec_context failed.\n");
+        return -1;
+    }
+    
+    if (output_token.length == 0)
+        return -1;
+
+    sess->gssapi_token = ne_base64(output_token.value, output_token.length);
+
+    NE_DEBUG(NE_DBG_HTTPAUTH, 
+             "Base64 encoded GSSAPI challenge: %s.\n", sess->gssapi_token);
+    sess->scheme = auth_scheme_gssapi;
+    return 0;
+}
+#endif
+
 /* Examine a digest challenge: return 0 if it is a valid Digest challenge,
  * else non-zero. */
-static int digest_challenge(auth_session *sess,
-			    struct auth_challenge *parms) 
+static int digest_challenge(auth_session *sess, struct auth_challenge *parms) 
 {
     struct ne_md5_ctx tmp;
     unsigned char tmp_md5[16];
-    char *password;
+    char password[NE_ABUFSIZ];
 
     /* Verify they've given us the right bits. */
     if (parms->alg == auth_alg_unknown ||
@@ -397,7 +400,7 @@ static int digest_challenge(auth_session *sess,
 	sess->realm = ne_strdup(parms->realm);
 
 	/* Not a stale response: really need user authentication */
-	if (get_credentials(sess, &password)) {
+	if (get_credentials(sess, password)) {
 	    /* Failed to get credentials */
 	    return -1;
 	}
@@ -406,14 +409,7 @@ static int digest_challenge(auth_session *sess,
     sess->scheme = auth_scheme_digest;
     sess->nonce = ne_strdup(parms->nonce);
     sess->cnonce = get_cnonce();
-    if (parms->domain) {
-	if (parse_domain(sess, parms->domain)) {
-	    /* TODO: Handle the error? */
-	}
-    } else {
-	sess->domain = NULL;
-	sess->domain_count = 0;
-    }
+    /* TODO: add domain handling. */
     if (parms->opaque != NULL) {
 	sess->opaque = ne_strdup(parms->opaque); /* don't strip the quotes */
     }
@@ -442,8 +438,8 @@ static int digest_challenge(auth_session *sess,
 	ne_md5_process_bytes(":", 1, &tmp);
 	ne_md5_process_bytes(sess->realm, strlen(sess->realm), &tmp);
 	ne_md5_process_bytes(":", 1, &tmp);
-	if (password != NULL)
-	    ne_md5_process_bytes(password, strlen(password), &tmp);
+	ne_md5_process_bytes(password, strlen(password), &tmp);
+	memset(password, 0, sizeof password); /* done with that. */
 	ne_md5_finish_ctx(&tmp, tmp_md5);
 	if (sess->alg == auth_alg_md5_sess) {
 	    unsigned char a1_md5[16];
@@ -467,7 +463,6 @@ static int digest_challenge(auth_session *sess,
 	    NE_DEBUG(NE_DBG_HTTPAUTH, "H(A1) is [%s]\n", sess->h_a1);
 	}
 	
-	NE_FREE(password);
     }
     
     NE_DEBUG(NE_DBG_HTTPAUTH, "I like this Digest challenge.\n");
@@ -497,11 +492,12 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
     /* Increase the nonce-count */
     if (sess->qop != auth_qop_none) {
 	sess->nonce_count++;
-	snprintf(nc_value, 9, "%08x", sess->nonce_count);
+	ne_snprintf(nc_value, 9, "%08x", sess->nonce_count);
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Nonce count is %d, nc is [%s]\n", 
-	       sess->nonce_count, nc_value);
+		 sess->nonce_count, nc_value);
     }
-    qop_value = qop_values[sess->qop];
+
+    qop_value = sess->qop == auth_qop_auth_int ? "auth-int" : "auth";
 
     /* Calculate H(A2). */
     ne_md5_init_ctx(&a2);
@@ -553,7 +549,7 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
 	 *    nc-value ":" unq(cnonce-value) ":" unq(qop-value) ":"
 	 */
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Have qop directive, digesting: [%s:%s:%s]\n",
-	       nc_value, sess->cnonce, qop_value);
+		 nc_value, sess->cnonce, qop_value);
 	ne_md5_process_bytes(nc_value, 8, &rdig);
 	ne_md5_process_bytes(":", 1, &rdig);
 	ne_md5_process_bytes(sess->cnonce, strlen(sess->cnonce), &rdig);
@@ -576,12 +572,13 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
     ret = ne_buffer_create();
 
     ne_buffer_concat(ret, 
-		   "Digest username=\"", sess->username, "\", "
-		   "realm=\"", sess->realm, "\", "
-		   "nonce=\"", sess->nonce, "\", "
-		   "uri=\"", req->uri, "\", "
-		   "response=\"", rdig_md5_ascii, "\", "
-		   "algorithm=\"", algorithm_names[sess->alg], "\"", NULL);
+		     "Digest username=\"", sess->username, "\", "
+		     "realm=\"", sess->realm, "\", "
+		     "nonce=\"", sess->nonce, "\", "
+		     "uri=\"", req->uri, "\", "
+		     "response=\"", rdig_md5_ascii, "\", "
+		     "algorithm=\"", sess->alg == auth_alg_md5 ? "MD5" : "MD5-sess", "\"", 
+		     NULL);
     
     if (sess->opaque != NULL) {
 	ne_buffer_concat(ret, ", opaque=\"", sess->opaque, "\"", NULL);
@@ -589,10 +586,9 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
 
     if (sess->qop != auth_qop_none) {
 	/* Add in cnonce and nc-value fields */
-	ne_buffer_concat(ret, 
-		       ", cnonce=\"", sess->cnonce, "\", "
-		       "nc=", nc_value, ", "
-		       "qop=\"", qop_values[sess->qop], "\"", NULL);
+	ne_buffer_concat(ret, ", cnonce=\"", sess->cnonce, "\", "
+			 "nc=", nc_value, ", "
+			 "qop=\"", qop_value, "\"", NULL);
     }
 
     ne_buffer_zappend(ret, "\r\n");
@@ -606,8 +602,7 @@ static char *request_digest(auth_session *sess, struct auth_request *req)
  * then also return a leading space-separated token, as *value == NULL.
  * Otherwise, if return value is 0, *key and *value will be non-NULL.
  * If return value is non-zero, parsing has ended. */
-static int tokenize(char **hdr, char **key, char **value, 
-		    int ischall)
+static int tokenize(char **hdr, char **key, char **value, int ischall)
 {
     char *pnt = *hdr;
     enum { BEFORE_EQ, AFTER_EQ, AFTER_EQ_QUOTED } state = BEFORE_EQ;
@@ -652,6 +647,10 @@ static int tokenize(char **hdr, char **key, char **value,
 	}
     } while (*++pnt != '\0');
     
+    if (state == BEFORE_EQ && ischall && *key != NULL) {
+	*value = NULL;
+    }
+
     *hdr = pnt;
 
     /* End of string: */
@@ -664,7 +663,8 @@ static int tokenize(char **hdr, char **key, char **value,
  *    0 if it gives a valid authentication for the server 
  *    non-zero otherwise (don't believe the response in this case!).
  */
-int verify_response(struct auth_request *req, auth_session *sess, const char *value) 
+static int verify_response(struct auth_request *req, auth_session *sess,
+			   const char *value) 
 {
     char *hdr, *pnt, *key, *val;
     auth_qop qop = auth_qop_none;
@@ -681,7 +681,8 @@ int verify_response(struct auth_request *req, auth_session *sess, const char *va
     }
     
     if (sess->scheme != auth_scheme_digest) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Found Auth-Info header not in response to Digest credentials - dodgy.\n");
+	NE_DEBUG(NE_DBG_HTTPAUTH, "Found Auth-Info header not in response "
+		 " to Digest credentials - dodgy.\n");
 	return -1;
     }
 
@@ -765,9 +766,9 @@ int verify_response(struct auth_request *req, auth_session *sess, const char *va
 		if (qop != auth_qop_none) {
 		    /* Add in qop-value */
 		    NE_DEBUG(NE_DBG_HTTPAUTH, "Digesting qop-value [%s:].\n", 
-			   qop_value);
+			     qop_value);
 		    ne_md5_process_bytes(qop_value, strlen(qop_value), 
-				       &sess->stored_rdig);
+					 &sess->stored_rdig);
 		    ne_md5_process_bytes(":", 1, &sess->stored_rdig);
 		}
 		/* Digest ":" H(A2) */
@@ -776,10 +777,10 @@ int verify_response(struct auth_request *req, auth_session *sess, const char *va
 		ne_md5_finish_ctx(&sess->stored_rdig, rdig_md5);
 		ne_md5_to_ascii(rdig_md5, rdig_md5_ascii);
 
-		NE_DEBUG(NE_DBG_HTTPAUTH, "Calculated response-digest of: [%s]\n",
-		       rdig_md5_ascii);
-		NE_DEBUG(NE_DBG_HTTPAUTH, "Given response-digest of:      [%s]\n",
-		       rspauth);
+		NE_DEBUG(NE_DBG_HTTPAUTH, "Calculated response-digest of: "
+			 "[%s]\n", rdig_md5_ascii);
+		NE_DEBUG(NE_DBG_HTTPAUTH, "Given response-digest of:      "
+			 "[%s]\n", rspauth);
 
 		/* And... do they match? */
 		okay = (strcasecmp(rdig_md5_ascii, rspauth) == 0)?0:-1;
@@ -795,18 +796,19 @@ int verify_response(struct auth_request *req, auth_session *sess, const char *va
     if (nextnonce != NULL) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Found nextnonce of [%s].\n", nextnonce);
 	if (sess->nonce != NULL)
-	    free(sess->nonce);
+	    ne_free(sess->nonce);
 	sess->nonce = ne_strdup(nextnonce);
     }
 
-    free(hdr);
+    ne_free(hdr);
 
     return okay;
 }
 
-
-/* A new challenge presented by the server */
-int auth_challenge(auth_session *sess, const char *value) 
+/* Passed the value of a "(Proxy,WWW)-Authenticate: " header field.
+ * Returns 0 if valid challenge was accepted; non-zero if no valid
+ * challenge was found. */
+static int auth_challenge(auth_session *sess, const char *value) 
 {
     char *pnt, *key, *val, *hdr;
     struct auth_challenge *chall = NULL, *challenges = NULL;
@@ -837,9 +839,14 @@ int auth_challenge(auth_session *sess, const char *value)
 	    } else if (strcasecmp(key, "digest") == 0) {
 		NE_DEBUG(NE_DBG_HTTPAUTH, "Digest scheme.\n");
 		chall->scheme = auth_scheme_digest;
+#ifdef HAVE_GSSAPI		
+	    } else if (strcasecmp(key, "gss-negotiate") == 0) {
+		NE_DEBUG(NE_DBG_HTTPAUTH, "GSSAPI scheme.\n");
+		chall->scheme = auth_scheme_gssapi;
+#endif
 	    } else {
 		NE_DEBUG(NE_DBG_HTTPAUTH, "Unknown scheme.\n");
-		free(chall);
+		ne_free(chall);
 		challenges = NULL;
 		break;
 	    }
@@ -862,8 +869,6 @@ int auth_challenge(auth_session *sess, const char *value)
 	    chall->nonce = val;
 	} else if (strcasecmp(key, "opaque") == 0) {
 	    chall->opaque = val;
-	} else if (strcasecmp(key, "domain") == 0) {
-	    chall->domain = val;
 	} else if (strcasecmp(key, "stale") == 0) {
 	    /* Truth value */
 	    chall->stale = (strcasecmp(val, "true") == 0);
@@ -875,18 +880,17 @@ int auth_challenge(auth_session *sess, const char *value)
 	    } else {
 		chall->alg = auth_alg_unknown;
 	    }
-	} else if (strcasecmp(val, "qop") == 0) {
-	    char **qops;
-	    int qop;
-	    qops = split_string(val, ',', NULL, " \r\n\t");
-	    chall->got_qop = 1;
-	    for (qop = 0; qops[qop] != NULL; qop++) {
-		if (strcasecmp(qops[qop], "auth") == 0) {
-		    chall->qop_auth = 1;
-		} else if (strcasecmp(qops[qop], "auth-int") == 0) {
-		    chall->qop_auth_int = 1;
-		}
-	    }
+	} else if (strcasecmp(key, "qop") == 0) {
+            /* iterate over each token in the value */
+            do {
+                const char *tok = ne_shave(ne_token(&val, ','), " \t");
+                
+                if (strcasecmp(tok, "auth") == 0) {
+                    chall->qop_auth = 1;
+                } else if (strcasecmp(tok, "auth-int") == 0 ) {
+                    chall->qop_auth_int = 1;
+                }
+            } while (val);
 	}
     }
 
@@ -894,26 +898,41 @@ int auth_challenge(auth_session *sess, const char *value)
 
     /* Did we find any challenges */
     if (challenges == NULL) {
-	free(hdr);
+	ne_free(hdr);
 	return -1;
     }
     
     success = 0;
 
-    NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for Digest challenges.\n");
-
-    /* Try a digest challenge */
+#ifdef HAVE_GSSAPI
+    NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for GSSAPI.\n");
+    /* Try a GSSAPI challenge */
     for (chall = challenges; chall != NULL; chall = chall->next) {
-	if (chall->scheme == auth_scheme_digest) {
-	    if (!digest_challenge(sess, chall)) {
+	if (chall->scheme == auth_scheme_gssapi) {
+	    if (!gssapi_challenge(sess, chall)) {
 		success = 1;
 		break;
 	    }
 	}
     }
+#endif
+
+    /* Try a digest challenge */
+    if (!success) {
+        NE_DEBUG(NE_DBG_HTTPAUTH, "Looking for Digest challenges.\n");
+        for (chall = challenges; chall != NULL; chall = chall->next) {
+        	if (chall->scheme == auth_scheme_digest) {
+        	    if (!digest_challenge(sess, chall)) {
+        		success = 1;
+        		break;
+        	    }
+        	}
+        }
+    }
 
     if (!success) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "No good Digest challenges, looking for Basic.\n");
+	NE_DEBUG(NE_DBG_HTTPAUTH, 
+		 "No good Digest challenges, looking for Basic.\n");
 	for (chall = challenges; chall != NULL; chall = chall->next) {
 	    if (chall->scheme == auth_scheme_basic) {
 		if (!basic_challenge(sess, chall)) {
@@ -935,11 +954,11 @@ int auth_challenge(auth_session *sess, const char *value)
 
     while (challenges != NULL) {
 	chall = challenges->next;
-	free(challenges);
+	ne_free(challenges);
 	challenges = chall;
     }
 
-    free(hdr);
+    ne_free(hdr);
 
     return !success;
 }
@@ -948,48 +967,44 @@ int auth_challenge(auth_session *sess, const char *value)
 static void auth_body_reader(void *cookie, const char *block, size_t length)
 {
     struct ne_md5_ctx *ctx = cookie;
-    NE_DEBUG(NE_DBG_HTTPAUTH, "Digesting %d bytes of response body.\n", length);
+    NE_DEBUG(NE_DBG_HTTPAUTH, 
+	     "Digesting %" NE_FMT_SIZE_T " bytes of response body.\n", length);
     ne_md5_process_bytes(block, length, ctx);
 }
 
-static void *ah_create(void *session, ne_request *request, const char *method,
-		       const char *uri)
+static void ah_create(ne_request *req, void *session, const char *method,
+		      const char *uri)
 {
     auth_session *sess = session;
     struct auth_request *areq = ne_calloc(sizeof *areq);
 
-    NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->resp_hdr);
+    NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->spec->resp_hdr);
 
-    areq->session = sess;
     areq->method = method;
     areq->uri = uri;
-    areq->request = request;
+    areq->request = req;
 
-    ne_add_response_header_handler(request, sess->resp_hdr,
+    ne_add_response_header_handler(req, sess->spec->resp_hdr,
 				   ne_duplicate_header, &areq->auth_hdr);
 
 	
-    ne_add_response_header_handler(request, sess->resp_info_hdr,
+    ne_add_response_header_handler(req, sess->spec->resp_info_hdr,
 				   ne_duplicate_header, 
 				   &areq->auth_info_hdr);
     
-    return areq;
+    sess->attempt = 0;
+
+    ne_set_request_private(req, sess->spec->id, areq);
 }
 
 
-static void ah_pre_send(void *cookie, ne_buffer *request)
+static void ah_pre_send(ne_request *r, void *cookie, ne_buffer *request)
 {
-    struct auth_request *req = cookie;
-    auth_session *sess = req->session;
-
+    auth_session *sess = cookie;
+    struct auth_request *req = ne_get_request_private(r, sess->spec->id);
 
     if (!sess->can_handle) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Not handling session.\n");
-    } else if (!is_in_domain(sess, req->uri)) {
-	/* We have moved out of the authentication domain */
-	NE_DEBUG(NE_DBG_HTTPAUTH, 
-	      "URI %s outside session domain, not handling.\n", req->uri);
-	req->will_handle = 0;
     } else {
 	char *value;
 
@@ -1011,50 +1026,51 @@ static void ah_pre_send(void *cookie, ne_buffer *request)
 	case auth_scheme_digest:
 	    value = request_digest(sess, req);
 	    break;
+#ifdef HAVE_GSSAPI	    
+	case auth_scheme_gssapi:
+	    value = request_gssapi(sess);
+	    break;
+#endif
 	default:
 	    value = NULL;
 	    break;
 	}
 
 	if (value != NULL) {
-	    ne_buffer_concat(request, sess->req_hdr, ": ", value, NULL);
-	    free(value);
+	    ne_buffer_concat(request, sess->spec->req_hdr, ": ", value, NULL);
+	    ne_free(value);
 	}
 
-	/* increase counter so we don't retry this >1. */
-	req->tries++;
-
     }
 
 }
 
 #define SAFELY(x) ((x) != NULL?(x):"null")
 
-static int ah_post_send(void *cookie, const ne_status *status)
+static int ah_post_send(ne_request *req, void *cookie, const ne_status *status)
 {
-    struct auth_request *areq = cookie;
-    auth_session *sess = areq->session;
+    auth_session *sess = cookie;
+    struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
     int ret = NE_OK;
 
     NE_DEBUG(NE_DBG_HTTPAUTH, 
-	  "ah_post_send (#%d), code is %d (want %d), %s is %s\n",
-	  areq->tries, status->code, sess->status_code, 
-	  SAFELY(sess->resp_hdr), SAFELY(areq->auth_hdr));
+	     "ah_post_send (#%d), code is %d (want %d), %s is %s\n",
+	     sess->attempt, status->code, sess->spec->status_code, 
+	     sess->spec->resp_hdr, SAFELY(areq->auth_hdr));
     if (areq->auth_info_hdr != NULL && 
 	verify_response(areq, sess, areq->auth_info_hdr)) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Response authentication invalid.\n");
-	ne_set_error(sess->sess, sess->fail_msg);
+	ne_set_error(sess->sess, _(sess->spec->fail_msg));
 	ret = NE_ERROR;
-    } else if (status->code == sess->status_code && 
-	       areq->auth_hdr != NULL && areq->tries == 0) {
+    } else if (status->code == sess->spec->status_code && 
+	       areq->auth_hdr != NULL) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Got challenge with code %d.\n", status->code);
 	if (!auth_challenge(sess, areq->auth_hdr)) {
 	    ret = NE_RETRY;
+	} else {
+	    clean_session(sess);
+	    ret = sess->spec->fail_code;
 	}
-    } else if (areq->tries > 0 && sess->status_code == status->code) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Authentication failed - bad password?\n");
-	clean_session(sess);
-	ret = sess->fail_code;
     }
 
     NE_FREE(areq->auth_info_hdr);
@@ -1063,70 +1079,58 @@ static int ah_post_send(void *cookie, const ne_status *status)
     return ret;
 }
 
-static void ah_destroy(void *cookie)
+static void ah_destroy(ne_request *req, void *session)
 {
-    free(cookie);
+    auth_session *sess = session;
+    struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
+    ne_free(areq);
 }
 
-static const ne_request_hooks ah_server_hooks = {
-    HOOK_SERVER_ID,
-    ah_create,
-    ah_pre_send,
-    ah_post_send,
-    ah_destroy
-};
-
-static const ne_request_hooks ah_proxy_hooks = {
-    HOOK_PROXY_ID,
-    ah_create,
-    ah_pre_send,
-    ah_post_send,
-    ah_destroy
-};
-
 static void free_auth(void *cookie)
 {
     auth_session *sess = cookie;
 
     clean_session(sess);
-    free(sess);
+    ne_free(sess);
 }
 
-void ne_set_server_auth(ne_session *sess, ne_request_auth callback, 
-			  void *userdata)
+static void auth_register(ne_session *sess,
+			  const struct auth_class *ahc, const char *id, 
+			  ne_auth_creds creds, void *userdata) 
 {
-    auth_session *auth_sess = auth_create(sess, callback, userdata);
+    auth_session *ahs = ne_calloc(sizeof *ahs);
 
-    /* Server auth details */
-    auth_sess->status_code = 401;
-    auth_sess->fail_code = NE_AUTH;
-    auth_sess->resp_hdr = "WWW-Authenticate";
-    auth_sess->resp_info_hdr = "Authentication-Info";
-    auth_sess->req_hdr = "Authorization";
-    auth_sess->fail_msg = _("Server was not authenticated correctly.");
+    ahs->creds = creds;
+    ahs->userdata = userdata;
+    ahs->sess = sess;
+    ahs->spec = ahc;
 
-    ne_add_hooks(sess, &ah_server_hooks, auth_sess, free_auth);
+    /* Register hooks */
+    ne_hook_create_request(sess, ah_create, ahs);
+    ne_hook_pre_send(sess, ah_pre_send, ahs);
+    ne_hook_post_send(sess, ah_post_send, ahs);
+    ne_hook_destroy_request(sess, ah_destroy, ahs);
+    ne_hook_destroy_session(sess, free_auth, ahs);
+
+    ne_set_session_private(sess, id, ahs);
 }
 
-void ne_set_proxy_auth(ne_session *sess, ne_request_auth callback, 
-			 void *userdata)
+void ne_set_server_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
 {
-    auth_session *auth_sess = auth_create(sess, callback, userdata);
+    auth_register(sess, &ah_server_class, HOOK_SERVER_ID, creds, userdata);
+}
 
-    /* Proxy auth details */
-    auth_sess->status_code = 407;
-    auth_sess->fail_code = NE_PROXYAUTH;
-    auth_sess->resp_hdr = "Proxy-Authenticate";
-    auth_sess->resp_info_hdr = "Proxy-Authentication-Info";
-    auth_sess->req_hdr = "Proxy-Authorization";
-    auth_sess->fail_msg = _("Proxy server was not authenticated correctly.");
-
-    ne_add_hooks(sess, &ah_proxy_hooks, auth_sess, free_auth);
+void ne_set_proxy_auth(ne_session *sess, ne_auth_creds creds, void *userdata)
+{
+    auth_register(sess, &ah_proxy_class, HOOK_PROXY_ID, creds, userdata);
 }
 
 void ne_forget_auth(ne_session *sess)
 {
-    clean_session(ne_session_hook_private(sess, HOOK_SERVER_ID));
-    clean_session(ne_session_hook_private(sess, HOOK_PROXY_ID));
+    auth_session *as;
+    if ((as = ne_get_session_private(sess, HOOK_SERVER_ID)) != NULL)
+	clean_session(as);
+    if ((as = ne_get_session_private(sess, HOOK_PROXY_ID)) != NULL)
+	clean_session(as);
 }
 
