@@ -1,6 +1,6 @@
 /* 
    HTTP request/response handling
-   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -91,12 +91,9 @@ struct body_reader {
 };
 
 struct ne_request_s {
-    const char *method;
-    char *uri; /* Request-URI */
-    
-    /*** Request ***/
+    char *method, *uri; /* method and Request-URI */
 
-    ne_buffer *headers;
+    ne_buffer *headers; /* request headers */
 
     /* Request body. */
     ne_provide_body body_cb;
@@ -217,18 +214,14 @@ static int aborted(ne_request *req, const char *doing, ssize_t code)
 	ne_set_error(sess, _("%s: connection timed out."), doing);
 	ret = NE_TIMEOUT;
 	break;
+    case NE_SOCK_ERROR:
+    case NE_SOCK_RESET:
+    case NE_SOCK_TRUNC:
+        ne_set_error(sess, "%s: %s", doing, ne_sock_error(sess->socket));
+        break;
     case 0:
 	ne_set_error(sess, "%s", doing);
 	break;
-    default:
-	if (sess->socket != NULL) {
-	    const char *err = ne_sock_error(sess->socket);
-	    ne_set_error(sess, "%s: %s", doing, err);
-	} else {
-	    char err[200];
-	    ne_strerror(errno, err, sizeof err);
-	    ne_set_error(sess, "%s: %s", doing, err);
-	}
     }
 
     ne_close_connection(sess);
@@ -438,23 +431,23 @@ static int send_request_body(ne_request *req)
 static void add_fixed_headers(ne_request *req) 
 {
     if (req->session->user_agent) {
-	ne_buffer_concat(req->headers, 
-			"User-Agent: ", req->session->user_agent, EOL, NULL);
+        ne_buffer_zappend(req->headers, req->session->user_agent);
     }
-    /* Send Connection: Keep-Alive for pre-1.1 origin servers, so we
-     * might get a persistent connection. 2068 sec 19.7.1 says we MUST
-     * NOT do this for proxies, though. So we don't.  Note that on the
-     * first request on any session, we don't know whether the server
-     * is 1.1 compliant, so we presume that it is not. */
+
+    /* Send Connection: Keep-Alive to pre-1.1 origin servers to try
+     * harder to get a persistent connection, except if using a proxy
+     * as per 2068 sec 19.7.1.  Always add TE: trailers since those
+     * are understood. */
     if (!req->session->is_http11 && !req->session->use_proxy) {
-	ne_buffer_zappend(req->headers, "Keep-Alive: " EOL);
-	ne_buffer_zappend(req->headers, "Connection: TE, Keep-Alive" EOL);
+	ne_buffer_zappend(req->headers, 
+                          "Keep-Alive: " EOL
+                          "Connection: TE, Keep-Alive" EOL
+                          "TE: trailers" EOL);
     } else {
-	ne_buffer_zappend(req->headers, "Connection: TE" EOL);
+	ne_buffer_zappend(req->headers, 
+                          "Connection: TE" EOL
+                          "TE: trailers" EOL);
     }
-    /* We send TE: trailers since we understand trailers in the chunked
-     * response. */
-    ne_buffer_zappend(req->headers, "TE: trailers" EOL);
 
 }
 
@@ -468,12 +461,14 @@ int ne_accept_2xx(void *userdata, ne_request *req, const ne_status *st)
     return (st->klass == 2);
 }
 
-/* Handler for the "Transfer-Encoding" response header */
+/* Handler for the "Transfer-Encoding" response header: treat *any*
+ * such header as implying a chunked response, per the "Protocol
+ * Compliance" statement in the manual. */
 static void te_hdr_handler(void *userdata, const char *value) 
 {
     struct ne_response *resp = userdata;
-    if (strcasecmp(value, "chunked") == 0)
-	resp->mode = R_CHUNKED;
+
+    resp->mode = R_CHUNKED;    
 }
 
 /* Handler for the "Connection" response header */
@@ -511,8 +506,8 @@ ne_request *ne_request_create(ne_session *sess,
     add_fixed_headers(req);
 
     /* Set the standard stuff */
-    req->method = method;
-    req->method_is_head = (strcmp(req->method, "HEAD") == 0);
+    req->method = ne_strdup(method);
+    req->method_is_head = (strcmp(method, "HEAD") == 0);
 
     /* Add in handlers for all the standard HTTP headers. */
     ne_add_response_header_handler(req, "Content-Length", 
@@ -537,7 +532,7 @@ ne_request *ne_request_create(ne_session *sess,
 
 	for (hk = sess->create_req_hooks; hk != NULL; hk = hk->next) {
 	    ne_create_request_fn fn = (ne_create_request_fn)hk->fn;
-	    fn(req, hk->userdata, method, path);
+	    fn(req, hk->userdata, method, req->uri);
 	}
     }
 
@@ -650,24 +645,25 @@ void ne_request_destroy(ne_request *req)
     struct hook *hk, *next_hk;
     int n;
 
-    NE_FREE(req->uri);
+    ne_free(req->uri);
+    ne_free(req->method);
 
     for (rdr = req->body_readers; rdr != NULL; rdr = next_rdr) {
 	next_rdr = rdr->next;
-	free(rdr);
+	ne_free(rdr);
     }
 
     for (hdlr = req->header_catchers; hdlr != NULL; hdlr = next_hdlr) {
 	next_hdlr = hdlr->next;
-	free(hdlr);
+	ne_free(hdlr);
     }
 
     for (n = 0; n < HH_HASHSIZE; n++) {
 	for (hdlr = req->header_handlers[n]; hdlr != NULL; 
 	     hdlr = next_hdlr) {
 	    next_hdlr = hdlr->next;
-	    free(hdlr->name);
-	    free(hdlr);
+	    ne_free(hdlr->name);
+	    ne_free(hdlr);
 	}
     }
 
@@ -681,14 +677,14 @@ void ne_request_destroy(ne_request *req)
 
     for (hk = req->private; hk; hk = next_hk) {
 	next_hk = hk->next;
-	free(hk);
+	ne_free(hk);
     }
 
     if (req->status.reason_phrase)
-	free(req->status.reason_phrase);
+	ne_free(req->status.reason_phrase);
 
     NE_DEBUG(NE_DBG_HTTP, "Request ends.\n");
-    free(req);
+    ne_free(req);
 }
 
 
@@ -765,6 +761,7 @@ static int read_response_block(ne_request *req, struct ne_response *resp,
 	(readlen == NE_SOCK_CLOSED ||
 	 (readlen == NE_SOCK_TRUNC && resp->total == 0))) {
 	NE_DEBUG(NE_DBG_HTTP, "Got EOF.\n");
+	req->can_persist = 0;
 	readlen = 0;
     } else if (readlen < 0) {
 	return aborted(req, _("Could not read response body"), readlen);
@@ -782,10 +779,11 @@ static int read_response_block(ne_request *req, struct ne_response *resp,
 	    char crlfbuf[2];
 	    /* If we've read a whole chunk, read a CRLF */
 	    readlen = ne_sock_fullread(sock, crlfbuf, 2);
-	    if (readlen < 0 || strncmp(crlfbuf, EOL, 2) != 0) {
-		return aborted(req, _("Error reading chunked response body"),
-			       readlen);
-	    }
+            if (readlen < 0)
+                return aborted(req, _("Could not read chunk delimiter"),
+                               readlen);
+            else if (crlfbuf[0] != '\r' || crlfbuf[1] != '\n')
+                return aborted(req, _("Chunk delimiter was invalid"), 0);
 	}
     } else if (resp->mode == R_CLENGTH) {
 	resp->left -= readlen;
@@ -822,9 +820,6 @@ static ne_buffer *build_request(ne_request *req)
     struct hook *hk;
     ne_buffer *buf = ne_buffer_create();
 
-    /* Reset the buffer. */
-    ne_buffer_clear(buf);
-
     /* Add Request-Line and Host header: */
     ne_buffer_concat(buf, req->method, " ", req->uri, " HTTP/1.1" EOL,
 		     "Host: ", req->session->server.hostport, EOL, NULL);
@@ -863,7 +858,7 @@ static void dump_request(const char *request)
 	    }
 	}
 	NE_DEBUG(NE_DBG_HTTP, "Sending request headers:\n%s", reqdebug);
-	free(reqdebug);
+	ne_free(reqdebug);
     }
 }
 
@@ -891,9 +886,8 @@ static inline void strip_eol(char *buf, ssize_t *len)
  * appropriate return value given a 'retry' flag, the socket error
  * 'code', and the return value 'acode' from the aborted() function. */
 #define RETRY_RET(retry, code, acode) \
-((((code) == NE_SOCK_CLOSED || (code) == NE_SOCK_RESET) && retry) \
-? NE_RETRY : (acode))
-/* TODO: probably need NE_SOCK_TRUNC in there too. */
+((((code) == NE_SOCK_CLOSED || (code) == NE_SOCK_RESET || \
+ (code) == NE_SOCK_TRUNC) && retry) ? NE_RETRY : (acode))
 
 /* Read and parse response status-line into 'status'.  'retry' is non-zero
  * if an NE_RETRY should be returned if an EOF is received. */
@@ -911,7 +905,7 @@ static int read_status_line(ne_request *req, ne_status *status, int retry)
     NE_DEBUG(NE_DBG_HTTP, "[status-line] < %s", buffer);
     strip_eol(buffer, &ret);
     
-    if (status->reason_phrase) free(status->reason_phrase);
+    if (status->reason_phrase) ne_free(status->reason_phrase);
     memset(status, 0, sizeof *status);
 
     if (ne_parse_statusline(buffer, status))
@@ -944,7 +938,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
 {
     ne_session *sess = req->session;
     ssize_t ret = NE_OK;
-    int aret, sentbody = 0, retry;
+    int sentbody = 0, retry;
     ne_status *status = &req->status;
 
     /* Send the Request-Line and headers */
@@ -958,7 +952,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
     ret = ne_sock_fullwrite(req->session->socket, request->data, 
 			    ne_buffer_size(request));
     if (ret < 0) {
-	aret = aborted(req, _("Could not send request"), ret);
+	int aret = aborted(req, _("Could not send request"), ret);
 	return RETRY_RET(retry, ret, aret);
     }
     
@@ -969,7 +963,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
 	/* Send request body, if not using 100-continue. */
 	ret = send_request_body(req);
 	if (ret < 0) {
-	    aret = aborted(req, _("Could not send request body"), ret);
+	    int aret = aborted(req, _("Could not send request body"), ret);
 	    return RETRY_RET(sess, ret, aret);
 	}
     }
@@ -1062,12 +1056,10 @@ static int read_message_header(ne_request *req, char *buf, size_t buflen)
 /* Apache's default is 100, seems reasonable. */
 #define MAX_HEADER_FIELDS 100
 
-#define MAX_HEADER_LENGTH 8192
-
 /* Read response headers.  Returns NE_* code, sets session error. */
 static int read_response_headers(ne_request *req) 
 {
-    char hdr[MAX_HEADER_LENGTH];
+    char hdr[8192]; /* max header length */
     int ret, count = 0;
     
     while ((ret = read_message_header(req, hdr, sizeof hdr)) == NE_RETRY 
@@ -1134,6 +1126,8 @@ static int lookup_host(ne_session *sess, struct host_info *info)
 	ne_set_error(sess, _("Could not resolve hostname `%s': %s"), 
 		     info->hostname,
 		     ne_addr_error(info->address, buf, sizeof buf));
+	ne_addr_destroy(info->address);
+	info->address = NULL;
 	return NE_LOOKUP;
     } else {
 	return NE_OK;
@@ -1171,25 +1165,26 @@ int ne_begin_request(ne_request *req)
     ne_buffer_destroy(data);
     if (ret != NE_OK) return ret;
 
-    if (st->major_version > 1 || 
-	(st->major_version == 1 && st->minor_version > 0))
-	/* Remember that server claims HTTP/1.1 compliance. */
-	req->session->is_http11 = 1;
+    /* Determine whether server claims HTTP/1.1 compliance. */
+    req->session->is_http11 = (st->major_version == 1 && 
+                               st->minor_version > 0) || st->major_version > 1;
+
+    /* Persistent connections supported implicitly in HTTP/1.1 */
+    if (req->session->is_http11) req->can_persist = 1;
 
     ne_set_error(req->session, "%d %s", st->code, st->reason_phrase);
 
-    /* persistent connections supported implicitly in HTTP/1.1 */
-    if (req->session->is_http11) req->can_persist = 1;
-
     /* Read the headers */
     HTTP_ERR(read_response_headers(req));
 
+#ifdef NEON_SSL
     /* Special case for CONNECT handling: the response has no body,
      * and the connection can persist. */
     if (req->session->in_connect && st->klass == 2) {
 	req->resp.mode = R_NO_BODY;
 	req->can_persist = 1;
     }
+#endif
 
     /* HEAD requests and 204, 304 responses have no response body,
      * regardless of what headers are present. */
@@ -1277,6 +1272,7 @@ ne_session *ne_get_session(const ne_request *req)
     return req->session;
 }
 
+#ifdef NEON_SSL
 /* Create a CONNECT tunnel through the proxy server.
  * Returns HTTP_* */
 static int proxy_tunnel(ne_session *sess)
@@ -1306,6 +1302,7 @@ static int proxy_tunnel(ne_session *sess)
     ne_request_destroy(req);
     return ret;
 }
+#endif
 
 /* Make new TCP connection to server at 'host' of type 'name'.  Note
  * that once a connection to a particular network address has
@@ -1317,6 +1314,12 @@ static int proxy_tunnel(ne_session *sess)
 static int do_connect(ne_request *req, struct host_info *host, const char *err)
 {
     ne_session *const sess = req->session;
+    int ret;
+
+    if ((sess->socket = ne_sock_create()) == NULL) {
+        ne_set_error(sess, _("Could not create socket"));
+        return NE_ERROR;
+    }
 
     if (host->current == NULL)
 	host->current = ne_addr_first(host->address);
@@ -1327,27 +1330,25 @@ static int do_connect(ne_request *req, struct host_info *host, const char *err)
 	if (ne_debug_mask & NE_DBG_HTTP) {
 	    char buf[150];
 	    NE_DEBUG(NE_DBG_HTTP, "Connecting to %s\n",
-		     ne_addr_print(host->current, buf, sizeof buf));
+		     ne_iaddr_print(host->current, buf, sizeof buf));
 	}
 #endif
-	sess->socket = ne_sock_connect(host->current, host->port);
-    } while (sess->socket == NULL && /* try the next address... */
+	ret = ne_sock_connect(sess->socket, host->current, host->port);
+    } while (ret && /* try the next address... */
 	     (host->current = ne_addr_next(host->address)) != NULL);
 
-    if (sess->socket == NULL) {
-	aborted(req, err, NE_SOCK_ERROR);
+    if (ret) {
+        ne_set_error(sess, "%s: %s", err, ne_sock_error(sess->socket));
+        ne_sock_close(sess->socket);
 	return NE_CONNECT;
     }
 
     notify_status(sess, ne_conn_connected, sess->proxy.hostport);
     
-    if (sess->progress_cb) {
-	ne_register_progress(sess->socket, sess->progress_cb,
-			       sess->progress_ud);
-    }
     if (sess->rdtimeout)
 	ne_sock_read_timeout(sess->socket, sess->rdtimeout);
 
+    sess->connected = 1;
     /* clear persistent connection flag. */
     sess->persisted = 0;
     return NE_OK;
@@ -1356,48 +1357,34 @@ static int do_connect(ne_request *req, struct host_info *host, const char *err)
 static int open_connection(ne_request *req) 
 {
     ne_session *sess = req->session;
+    int ret;
+    
+    if (sess->connected) return NE_OK;
 
-    if (req->session->use_proxy) {
-	switch(sess->connected) {
-	case 0:
-	    HTTP_ERR(do_connect(req, &sess->proxy,
-				_("Could not connect to proxy server")));
-	    sess->connected = 1;
-	    /* FALL-THROUGH */
-	case 1:
-	    if (sess->use_ssl && !sess->in_connect) {
-		int ret;
-		ret = proxy_tunnel(sess);
-		if (ret != NE_OK) {
-		    ne_close_connection(sess);
-		    return ret;
-		}
-		ret = ne_negotiate_ssl(req);
-		if (ret) {
-		    NE_DEBUG(NE_DBG_SOCKET, "Terminating connection.\n");
-		    ne_close_connection(sess);
-		    return ret;
-		}
-		sess->connected = 2;
-	    }
-	    break;
-	default:
-	    /* We've got everything we need */
-	    break;	    
-	}
-    } else if (sess->connected == 0) {
-	HTTP_ERR(do_connect(req, &sess->server, 
-			    _("Could not connect to server")));
-	if (sess->use_ssl) {
-	    int ret = ne_negotiate_ssl(req);
-	    if (ret) {
-		NE_DEBUG(NE_DBG_SOCKET, "Terminating connection.\n");
-		ne_sock_close(sess->socket);
-		return ret;
-	    }
-	}
+    if (!sess->use_proxy)
+	ret = do_connect(req, &sess->server, _("Could not connect to server"));
+    else
+	ret = do_connect(req, &sess->proxy,
+			 _("Could not connect to proxy server"));
 
-	sess->connected = 1;
+    if (ret != NE_OK) return ret;
+
+#ifdef NEON_SSL
+    /* Negotiate SSL layer if required. */
+    if (sess->use_ssl && !sess->in_connect) {
+        /* CONNECT tunnel */
+        if (req->session->use_proxy)
+            ret = proxy_tunnel(sess);
+        
+        if (ret == NE_OK)
+            ret = ne_negotiate_ssl(req);
+
+        /* This is probably only really needed for ne_negotiate_ssl
+         * failures as proxy_tunnel will fail via aborted(). */
+        if (ret != NE_OK)
+            ne_close_connection(sess);
     }
-    return NE_OK;
+#endif
+    
+    return ret;
 }
