@@ -91,6 +91,8 @@ char *proxy_hostname;
 int proxy_port;
 char *server_username = NULL, *server_password = NULL;
 
+ne_ssl_client_cert *client_cert;
+
 /* Global options */
 
 char *path, /* current working collection */
@@ -182,48 +184,33 @@ int set_path(const char *newpath)
     }
 }
 
-static int privkey_prompt(void *userdata, char *buf, size_t buflen)
-{
-    static char prompt[BUFSIZ];
-    const char *filename = userdata;
-    int len;
-
-    snprintf(prompt, BUFSIZ, _("Enter private key password for `%s': "), 
-	     filename);
-    
-    do {
-	char *pass = fm_getpassword(prompt);
-	if (pass == NULL) {
-	    return -1;
-	}
-	strncpy(buf, pass, buflen);
-	/* zero it out the buffer now */
-	memset(pass, 0, PASSWORDLEN);
-	len = strlen(buf);
-	if (len < 4) {
-	    printf(_("Private key password must be 4 or more characters!\n"));
-	}
-    } while (len < 4);
- 
-    return 0;
-}
-
 static int cert_verify(void *ud, int failures, const ne_ssl_certificate *c)
 {
-    char *tmp;
+    char *tmp, from[NE_SSL_VDATELEN], to[NE_SSL_VDATELEN];
+    const char *ident;
 
-    printf(_("WARNING: Untrusted server certificate presented:\n"));
-    if (failures & NE_SSL_CNMISMATCH) {
+    ident = ne_ssl_cert_identity(c);
+
+    if (ident)
+        printf(_("WARNING: Untrusted server certificate presented for `%s':\n"),
+               ident);
+    else
+        puts(_("WARNING: Untrusted server certificate presented:\n"));
+
+    if (failures & NE_SSL_IDMISMATCH) {
 	printf(_("Certificate was issued to hostname `%s' rather than `%s'\n"),
-	       c->subject->commonName, server.host);
+	       ne_ssl_cert_identity(c), server.host);
 	printf(_("This connection could have been intercepted.\n"));
     }
 
 #define PRINT_AND_FREE(str, dn) \
 tmp = ne_ssl_readable_dname(dn); printf(str, tmp); free(tmp)
 
-    PRINT_AND_FREE(_("Issued to: %s\n"), c->subject);
-    PRINT_AND_FREE(_("Issued by: %s\n"), c->subject);
+    PRINT_AND_FREE(_("Issued to: %s\n"), ne_ssl_cert_subject(c));
+    PRINT_AND_FREE(_("Issued by: %s\n"), ne_ssl_cert_issuer(c));
+
+    ne_ssl_cert_validity(c, from, to);
+    printf(_("Certificate is valid from %s to %s\n"), from, to);
 
     if (isatty(STDIN_FILENO)) {
 	printf(_("Do you wish to accept the certificate? (y/n) "));
@@ -234,32 +221,65 @@ tmp = ne_ssl_readable_dname(dn); printf(str, tmp); free(tmp)
     }
 }
 
+static void provide_clicert(void *userdata, ne_session *sess,
+                            const ne_ssl_dname *const *dname, int dncount)
+{
+    const char *ccfn = userdata;
+    int n;
+
+    printf("The server has requested a client certificate.\n");
+
+#if 0
+    /* display CA names? */
+    for (n = 0; n < dncount; n++) {
+        char *dn = ne_ssl_readable_dname(dname[n]);
+        printf("Name: %s\n", dn);
+        free(dn);
+    }
+#endif
+    
+    if (ne_ssl_clicert_encrypted(client_cert)) {
+        const char *name = ne_ssl_clicert_name(client_cert);
+        char *pass;
+        
+        if (!name) name = ccfn;
+        
+        printf("Client certificate `%s' is encrypted.\n", name);
+        
+        for (n = 0; n < 3; n++) {
+            pass = fm_getpassword(_("Decryption password: "));
+            if (pass == NULL) break;
+            if (ne_ssl_clicert_decrypt(client_cert, pass)) {
+                printf("Password incorrect, try again.\n");
+            } else {
+                break;
+            }
+        }
+    }
+    
+    if (!ne_ssl_clicert_encrypted(client_cert)) {
+        printf("Using client certificate.\n");
+        ne_ssl_set_clicert(session, client_cert);
+    }
+    
+}
+
 static int setup_ssl(void)
 {
-    char *cc = get_option(opt_cert);
-    char *key = get_option(opt_certkey);
+    char *ccfn = get_option(opt_clicert);
 
-    if (ne_ssl_load_default_ca(session)) {
-	printf(_("Warning: Could not load default CA list from SSL library.\n"
-		 "Check SSL library is installed correctly.\n"));
-    }
+    ne_ssl_trust_default_ca(session);
 	      
-
     ne_ssl_set_verify(session, cert_verify, NULL);
 
-    /* Set the client cert too if we have one. */
-    if (cc != NULL) {
-	
-	/* Use our private key password prompt. */
-	ne_ssl_keypw_prompt(session, privkey_prompt, cc);
-
-	if (ne_ssl_load_pem(session, cc, key)) {
-	    printf(_("Could not use client certificate `%s' (wrong password)?\n"),
-		   cc);
-	    return -1;
-	} else {
-	    printf(_("Using client certificate `%s'.\n"), cc);
-	}
+    if (ccfn) {
+        client_cert = ne_ssl_clicert_read(ccfn);
+        if (client_cert) {
+            ne_ssl_provide_clicert(session, provide_clicert, ccfn);
+        } else {
+            printf("Could not load client certificate from `%s'.\n",
+                   ccfn);
+        }
     }
 
     return 0;
@@ -362,7 +382,7 @@ void open_connection(const char *url)
     have_connection = false;
 
     ne_set_expect100(session, use_expect);
-    ne_set_useragent(session, PACKAGE "/" VERSION);
+    ne_set_useragent(session, "cadaver/" PACKAGE_VERSION);
     ne_set_server_auth(session, supply_creds_server, NULL);
     ne_set_proxy_auth(session, supply_creds_proxy, NULL);
     
@@ -430,7 +450,7 @@ static void parse_args(int argc, char **argv)
     while ((optc = getopt_long(argc, argv, "ehtp:V", opts, NULL)) != -1) {
 	switch (optc) {
 	case 'h': usage(); exit(-1);
-	case 'V': execute_version(); exit(-1);
+	case 'V': execute_about(); exit(-1);
 	case 'e': use_expect = true; break;
 	case 'p': set_proxy(optarg); break;
 	case 't': tolerant = 1; break;
@@ -460,8 +480,10 @@ static char * read_command(void)
 {
     static char prompt[BUFSIZ];
     if (path) {
-	snprintf(prompt, BUFSIZ, "dav:%s%c ", path,
+        char *p = ne_path_unescape(path);
+	snprintf(prompt, BUFSIZ, "dav:%s%c ", p,
 		  dav_collection?'>':'?');
+        ne_free(p);
     } else {
 	sprintf(prompt, "dav:!> ");
     }
@@ -793,8 +815,10 @@ static void init_options(void)
     set_option(opt_namespace, ne_strdup(DEFAULT_NAMESPACE));
     set_bool_option(opt_overwrite, 1);
     set_bool_option(opt_quiet, 1);
+    set_bool_option(opt_searchall, 1);
     lockdepth = NE_DEPTH_INFINITE;
     lockscope = ne_lockscope_exclusive;
+    searchdepth = NE_DEPTH_INFINITE;
 
     /* This is what Markus Kahn says we should do. */
     if ((tmp = getenv("LC_ALL")) ||
@@ -818,8 +842,8 @@ int main(int argc, char *argv[])
     progname = argv[0];
 
 #ifdef ENABLE_NLS
-    bindtextdomain(PACKAGE, LOCALEDIR);
-    textdomain(PACKAGE);
+    bindtextdomain(PACKAGE_NAME, LOCALEDIR);
+    textdomain(PACKAGE_NAME);
 #endif /* ENABLE_NLS */
 
     ne_debug_init(stderr, 0);
@@ -839,8 +863,8 @@ int main(int argc, char *argv[])
     init_netrc();
 
     init_signals();
-    init_rcfile();
     init_locking();
+    init_rcfile();
     
     parse_args(argc, argv);
 
@@ -1094,8 +1118,8 @@ static int supply_creds_server(void *userdata, const char *realm, int attempt,
 
     /* Pass back netrc values if we have them. */
     if (server_username != NULL && server_password != NULL) {
-	strncpy(username, server_username, NE_ABUFSIZ);
-	strncpy(password, server_password, NE_ABUFSIZ);
+	ne_strnzcpy(username, server_username, NE_ABUFSIZ);
+	ne_strnzcpy(password, server_password, NE_ABUFSIZ);
 	return 0;
     }
 
