@@ -51,18 +51,21 @@
 
 /* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
 enum command_id {
-    cmd_ls, cmd_cd, cmd_quit, cmd_open, cmd_close, cmd_version, cmd_pwd,
+    cmd_ls, cmd_cd, cmd_quit, cmd_open, cmd_close, cmd_about, cmd_pwd,
     cmd_help, cmd_put, cmd_get, cmd_mkcol, cmd_delete, cmd_move, cmd_copy, 
     cmd_less, cmd_cat, cmd_lpwd, cmd_lcd, cmd_lls, cmd_mput, cmd_mget,
     cmd_echo, cmd_set, cmd_unset, cmd_rmcol, cmd_lock, cmd_unlock,
     cmd_steal, cmd_discover, cmd_showlocks, cmd_propedit, cmd_propnames,
     cmd_propget, cmd_propset, cmd_propdel, cmd_chexec, cmd_edit, cmd_logout,
+    cmd_describe, cmd_search, cmd_version, cmd_checkin, cmd_checkout,
+    cmd_uncheckout, cmd_history, cmd_label,
     cmd_unknown
 /* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
 };
 
 struct command {
     enum command_id id; /* unique ID */
+    const char *name; /* real command name. */
     unsigned int needs_connection; /* true if only works when connected */
     int min_args, max_args; /* min and max # of arguments */
     enum { /* parameter scope */
@@ -127,6 +130,7 @@ struct resource {
     size_t size;
     time_t modtime;
     int is_executable;
+    int is_vcr;         /* Is version resource */
     char *error_reason; /* error string returned for this resource */
     int error_status; /* error status returned for this resource */
     struct resource *next;
@@ -138,6 +142,16 @@ void open_connection(const char *url);
 void execute_ls(const char *remote);
 void execute_edit(const char *remote);
 
+void execute_version(const char *remote);
+void execute_checkin(const char *remote);
+void execute_checkout(const char *remote);
+void execute_uncheckout(const char *remote);
+void execute_history(const char *remote);
+
+void execute_label(const char *res, const char *act, const char *value);
+
+void execute_search(int count, const char **args);
+
 void free_resource(struct resource *res);
 void free_resource_list(struct resource *res);
 
@@ -159,3 +173,4 @@ void output(enum output_type, const char *fmt, ...)
 ;
 
 #endif /* CADAVER_H */
+
