@@ -89,9 +89,12 @@ const static struct {
 #define C(n) { cmd_##n, #n }
     C(ls), C(cd), C(quit), C(open), C(logout), C(close), C(set), C(unset), 
     C(pwd), C(help), C(put), C(get), C(mkcol), C(delete), C(move), C(copy),
-    C(less), C(cat), C(lpwd), C(lcd), C(lls), C(echo), C(quit),
+    C(less), C(cat), C(lpwd), C(lcd), C(lls), C(echo), C(quit), C(about),
     C(mget), C(mput), C(rmcol), C(lock), C(unlock), C(discover), C(steal),
     C(chexec), C(showlocks), C(version), C(propget), C(propset), C(propdel),
+    C(describe), C(search),
+    C(version), C(checkin), C(checkout), C(uncheckout), C(history),
+    C(label), 
 #if 0
 C(propedit), 
 #endif
@@ -295,9 +298,7 @@ static char *utf8_decode(const char *str)
 }
 
 /* Resolve path, appending trailing slash if resource is a
- * collection. This function MAY make a request to the server.
- * TODO: caching, so it MAY not, too.
- */
+ * collection. */
 static char *true_path(const char *res)
 {
     char *full;
@@ -527,12 +528,11 @@ unlock_fail:
     ne_lock_destroy(lock);
 }
 
-static void execute_mkcol(const char *filename)
+static void execute_mkcol(const char *name)
 {
-    char *remote;
-    remote = resolve_path(path, filename, true);
-    dispatch("Creating", filename, ne_mkcol, remote);
-    free(remote);
+    char *uri = resolve_path(path, name, true);
+    dispatch(_("Creating"), name, ne_mkcol, uri);
+    free(uri);
 }
 
 static int all_iterator(void *userdata, const ne_propname *pname,
@@ -735,6 +735,12 @@ static void execute_rmcol(const char *filename)
     free(remote);
 }
 
+/* TODO: can do utf-8 handling here. */
+static char *escape_path(const char *p)
+{
+    return ne_path_escape(p);
+}
+
 /* Like resolve_path except more intelligent. */
 static char *clever_path(const char *p, const char *src, 
 			 const char *dest)
@@ -754,7 +760,9 @@ static char *clever_path(const char *p, const char *src,
 	/* Moving a file to a collection... the destination should
 	 * be the basename of file concated with the collection. */
 	char *tmp = resolve_path(p, dest, true);
-	ret = ne_concat(tmp, base_name(src), NULL);
+        char *enc = escape_path(base_name(src));
+	ret = ne_concat(tmp, enc, NULL);
+        free(enc);
 	free(tmp);
     } else {
 	ret = resolve_path(p, dest, false);
@@ -875,26 +883,23 @@ static void simple_copy(const char *src, const char *dest)
 		       NE_DEPTH_INFINITE, src, dest));
 }
 
-/* TODO: can do utf-8 handling here. */
-static char *escape_path(const char *p)
-{
-    return ne_path_escape(p);
-}
-
-char *resolve_path(const char *p, const char *filename, int isdir)
+/* Return full path of 'filename', relative to 'p'.  'p' must be
+ * already URI-escaped; 'filename' must not be.  If 'iscoll' is
+ * non-zero, the returned path will have a trailing slash. */
+char *resolve_path(const char *p, const char *filename, int iscoll)
 {
     char *ret, *pnt;
     if (*filename == '/') {
 	/* It's absolute */
 	ret = escape_path(filename);
     } else if (strcmp(filename, ".") == 0) {
-	ret = escape_path(p);
+	ret = ne_strdup(p);
     } else {
 	pnt = escape_path(filename);
 	ret = ne_concat(p, pnt, NULL);
 	free(pnt);
     }
-    if (isdir && ret[strlen(ret)-1] != '/') {
+    if (iscoll && ret[strlen(ret)-1] != '/') {
 	char *newret = ne_concat(ret, "/", NULL);
 	free(ret);
 	ret = newret;
@@ -953,8 +958,14 @@ static void execute_get(const char *remote, const char *local)
 	if (fd < 0) {
 	    output(o_finish, _("failed:\n%s\n"), strerror(errno));
 	} else {
-	    out_result(ne_get(session, real_remote, fd));
-	    /* FIXME */ close(fd);
+            int ret = ne_get(session, real_remote, fd);
+            if (close(fd) && ret == NE_OK) {
+                int errnum = errno;
+                ret = NE_ERROR;
+                ne_set_error(session, _("Could not write to file: %s"),
+                             strerror(errnum));
+            }
+	    out_result(ret);
 	}
     }
     free(real_remote);
@@ -1035,8 +1046,7 @@ static void multi_cat(int argc, const char *argv[])
 static void multi_mput(int argc, const char *argv[])
 {
     for(; argv[0] != NULL; argv++) {
-	char *remote;
-	remote = clever_path(path, argv[0], path);
+	char *remote = resolve_path(path, argv[0], false);
 	simple_put(argv[0], remote);
 	free(remote);
     }    
@@ -1178,15 +1188,25 @@ static void execute_cd(const char *newpath)
 
 static void display_help_message(void)
 {
-    int n;
-    printf("Commands: \n");
+    unsigned int n, pos = 0;
+
+    printf("Available commands: \n ");
     
-    for(n = 0; commands[n].id != cmd_unknown; n++) {
+    for (n = 0; commands[n].id != cmd_unknown; n++, pos=(pos+1)%7) {
+        printf("%-11s", commands[n].name);
+        if (pos == 6) printf("\n ");
+#if 0        
 	if (commands[n].call && commands[n].short_help) {
 	    printf(" %-26s %s\n", commands[n].call, _(commands[n].short_help));
 	}
+#endif
     }
 
+    if (pos == 6)
+        putchar('\r');
+    else
+        putchar('\n');
+
     printf(_("Aliases: rm=delete, mkdir=mkcol, mv=move, cp=copy, "
 	     "more=less, quit=exit=bye\n"));
 }
@@ -1197,6 +1217,7 @@ static void execute_help(const char *arg)
 	display_help_message();
     } else {
 	const struct command *cmd = get_command(arg);
+
 	if (cmd) {
 	    printf(" `%s'   %s\n", cmd->call, cmd->short_help);
 	    if (cmd->needs_connection) {
@@ -1217,14 +1238,17 @@ static void execute_echo(int count, const char **args)
     putchar('\n');
 }
 
-void execute_version(void)
+void execute_about(void)
 {
-    printf(PACKAGE " " VERSION "\n%s\n", ne_version_string());
+    printf("cadaver " PACKAGE_VERSION "\n%s\n", ne_version_string());
 #ifdef HAVE_LIBREADLINE
     printf("readline %s\n", rl_library_version);
 #endif
 }
 
+/* The T? macros are used to cast the function pointer into the
+ * command structure.  Using GCC extensions this is done in a
+ * type-safe way; for non-GCC< it's type-unsafe. */
 #if defined(__GNUC__)
 #define T0(x) {take0: x}
 #define T1(x) {take1: x}
@@ -1243,11 +1267,11 @@ void execute_version(void)
 
 /* C1: connected, 1-arg function C2: connected, 2-arg function
  * U0: disconnected, 0-arg function. */
-#define C1(x,c,h) { cmd_##x, true, 1, 1, parmscope_remote, T1(execute_##x),c,h }
-#define U0(x,h) { cmd_##x, false, 0, 0, parmscope_none, T0(execute_##x),#x,h }
-#define UO1(x,c,h) { cmd_##x, false, 0, 1, parmscope_none, T1(execute_##x),c,h }
-#define C2M(x,c,h) { cmd_##x, true, 2, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
-#define C1M(x,c,h) { cmd_##x, true, 1, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
+#define C1(x,c,h) { cmd_##x, #x, true, 1, 1, parmscope_remote, T1(execute_##x),c,h }
+#define U0(x,h) { cmd_##x, #x, false, 0, 0, parmscope_none, T0(execute_##x),#x,h }
+#define UO1(x,c,h) { cmd_##x, #x, false, 0, 1, parmscope_none, T1(execute_##x),c,h }
+#define C2M(x,c,h) { cmd_##x, #x, true, 2, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
+#define C1M(x,c,h) { cmd_##x, #x, true, 1, CMD_VARY, parmscope_remote, TV(multi_##x),c,h }
 
 /* commands[] is not static because it would mean adding a bunch of
  * prototypes for execute_* etc, and declaring this at the top of the
@@ -1256,17 +1280,17 @@ void execute_version(void)
 /* Separate structures for commands and command names. */
 /* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
 const struct command commands[] = {
-    { cmd_ls, true, 0, 1, parmscope_remote, T1(execute_ls), 
+    { cmd_ls, "ls", true, 0, 1, parmscope_remote, T1(execute_ls), 
       N_("ls [path]"), N_("List contents of current [or other] collection") },
     C1(cd, N_("cd path"), N_("Change to specified collection")),
-    { cmd_pwd, true, 0, 0, parmscope_none, T0(execute_pwd),
+    { cmd_pwd, "pwd", true, 0, 0, parmscope_none, T0(execute_pwd),
       "pwd", N_("Display name of current collection") },
-    { cmd_put, true, 1, 2, parmscope_none, T2(execute_put),
+    { cmd_put, "put", true, 1, 2, parmscope_none, T2(execute_put),
       N_("put local [remote]"), N_("Upload local file") },
-    { cmd_get, true, 1, 2, parmscope_none, T2(execute_get),
+    { cmd_get, "get", true, 1, 2, parmscope_none, T2(execute_get),
       N_("get remote [local]"), N_("Download remote resource") },
     C1M(mget, N_("mget remote..."), N_("Download many remote resources")),
-    { cmd_mput, true, 1, CMD_VARY, parmscope_local, TV(multi_mput), 
+    { cmd_mput, "mput", true, 1, CMD_VARY, parmscope_local, TV(multi_mput), 
       N_("mput local..."), N_("Upload many local files") },
     C1(edit, N_("edit resource"), N_("Edit given resource")),
     C1M(less, N_("less remote..."), N_("Display remote resource through pager")), 
@@ -1278,57 +1302,86 @@ const struct command commands[] = {
     C2M(move, N_("move source... dest"), N_("Move resource(s) from source to dest")),
     
 /* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND */
-    
+
+    /*** Locking commands ***/
+
     C1(lock, N_("lock resource"), N_("Lock given resource")),
     C1(unlock, N_("unlock resource"), N_("Unlock given resource")),
     C1(discover, N_("discover resource"), N_("Display lock information for resource")),
     C1(steal, N_("steal resource"), N_("Steal lock token for resource")),
-    { cmd_showlocks, true, 0, 0, parmscope_none, T0(execute_showlocks),
+    { cmd_showlocks, "showlocks", true, 0, 0, parmscope_none, T0(execute_showlocks),
       "showlocks", N_("Display list of owned locks") },
     
-#if 0
-    C1(propedit, "propedit resource", "Enter property editor for resource"),
+#ifdef ENABLE_DELTAV
+    /*** DeltaV commands ***/
+    C1(version, N_("verrsion resource"), N_("Place given resource under version control")),
+    C1(checkin, N_("checkin resource"), N_("Checkin given resource")),
+    C1(checkout, N_("checkout resource"), N_("Checkout given resource")),
+    C1(uncheckout, N_("uncheckin resource"), N_("Uncheckout given resource")),
+    C1(history, N_("history resource"), N_("Show version history of resource")),
+
+    { cmd_label, "label", true, 3, 3, parmscope_none, T3(execute_label),
+      N_("label res [add|set|remove] labelname"),
+      N_("Set/Del/Edit label on resource") },
 #endif
+
+    /*** Property handling ***/
     C1(propnames, "propnames res", "Names of properties defined on resource"),
 
-    { cmd_chexec, true, 2, 2, parmscope_none, T2(execute_chexec),
+    { cmd_chexec, "chexec", true, 2, 2, parmscope_none, T2(execute_chexec),
       N_("chexec [+|-] remote"), N_("Change isexecutable property of resource") },
     
-    { cmd_propget, true, 1, 2, parmscope_none, T2(execute_propget),
+    { cmd_propget, "propget", true, 1, 2, parmscope_none, T2(execute_propget),
       N_("propget res [propname]"), 
       N_("Retrieve properties of resource") },
-    { cmd_propdel, true, 2, 2, parmscope_none, T2(execute_propdel),
+    { cmd_propdel, "propdel", true, 2, 2, parmscope_none, T2(execute_propdel),
       N_("propdel res propname"), 
       N_("Delete property from resource") },
-    { cmd_propset, true, 3, 3, parmscope_none, T3(execute_propset),
+    { cmd_propset, "propset", true, 3, 3, parmscope_none, T3(execute_propset),
       N_("propset res propname value"),
       N_("Set property on resource") },
 
-    { cmd_set, false, 0, 2, parmscope_none, T2(execute_set), 
+#ifdef ENABLE_DASL
+    { cmd_search, "search", true, 1, CMD_VARY, parmscope_remote, TV(execute_search),
+      N_("search query"), 
+      N_("DASL Search resource in current collection\n\n"
+	 " Examples:\n"
+	 "   - search where content length is smaller than 100:\n"
+	 "      > search getcontentlength < 100\n" 
+         "   - search where author is Smith or Jones\n"
+         "      > search author = Smith or author = Jones\n" EOL
+	 " Available operators and keywords:\n"
+	 "     - and, or , (, ), =, <, >, <=, >=, like\n" EOL
+         " (See also variables searchdepth, searchorder, searchdorder)\n") },
+#endif
+    
+    { cmd_set, "set", false, 0, 2, parmscope_none, T2(execute_set), 
       N_("set [option] [value]"), N_("Set an option, or display options") },
-    { cmd_open, false, 1, 1, parmscope_none, T1(open_connection), 
+    { cmd_open, "open", false, 1, 1, parmscope_none, T1(open_connection), 
       "open URL", N_("Open connection to given URL") },
-    { cmd_close, true, 0, 0, parmscope_none, T0(close_connection), 
+    { cmd_close, "close", true, 0, 0, parmscope_none, T0(close_connection), 
       "close", N_("Close current connection") },
-    { cmd_echo, false, 1, CMD_VARY, parmscope_remote, TV(execute_echo), 
+    { cmd_echo, "echo", false, 1, CMD_VARY, parmscope_remote, TV(execute_echo), 
       "echo", NULL },
-    { cmd_quit, false, 0, 1, parmscope_none, T1(NULL), "quit",
+    { cmd_quit, "quit", false, 0, 1, parmscope_none, T1(NULL), "quit",
       N_("Exit program") },
     /* Unconnected operation, 1 mandatory argument */
-    { cmd_unset, false, 1, 2, parmscope_none, T2(execute_unset), 
+    { cmd_unset, "unset", false, 1, 2, parmscope_none, T2(execute_unset), 
       N_("unset [option] [value]"), N_("Unsets or clears value from option.") },
     /* Unconnected operation, 0 arguments */
     UO1(lcd, N_("lcd [directory]"), N_("Change local working directory")), 
-    { cmd_lls, false, 0, CMD_VARY, parmscope_local, TV(execute_lls), 
+    { cmd_lls, "lls", false, 0, CMD_VARY, parmscope_local, TV(execute_lls), 
       N_("lls [options]"), N_("Display local directory listing") },
     U0(lpwd, N_("Print local working directory")),
-    { cmd_logout, true, 0, 0, parmscope_none, T0(execute_logout), "logout",
+    { cmd_logout, "logout", true, 0, 0, parmscope_none, T0(execute_logout), "logout",
       N_("Logout of authentication session") },
     UO1(help, N_("help [command]"), N_("Display help message")), 
-    U0(version, NULL),
-    { cmd_unknown, 0 } /* end-of-list marker, DO NOT move */
+    { cmd_describe, "describe", false, 1, 1, parmscope_none, T1(execute_describe),
+      "describe option", N_("Describe an option variable") },
+    U0(about, N_("Information about this version of cadaver") ),
 
 /* DON'T FORGET TO ADD A NEW COMMAND ALIAS WHEN YOU ADD A NEW COMMAND. */
 
+    { cmd_unknown, 0 } /* end-of-list marker, DO NOT move */
 };    
 
